import logging

logger_manager = logging.getLogger('pipeline_engine.manager')
logger_decorators = logging.getLogger('pipeline_engine.decorators')
logger_parallelism = logging.getLogger('pipeline_engine.parallelism')
